__author__ = 'maartenbreddels'
import numpy as np
import vaex.vaexfast
from vaex.utils import filesize_format
import logging
import vaex.utils
total_bytes = 0

logger = logging.getLogger("vaex.grids")


def dog(grid, sigma1, sigma2):
    return gf(grid, sigma1) - gf(grid, sigma2)


def gf(grid, sigma, **kwargs):
    import scipy.ndimage
    return scipy.ndimage.gaussian_filter(grid, sigma=sigma, **kwargs)


functions = {}
functions["dog"] = dog
functions["gf"] = gf


def grid_average(scope, counts_name="counts", weighted_name="weighted"):

    counts = scope.evaluate(counts_name)
    weighted = scope.evaluate(weighted_name)
    logger.debug("evaluating average: counts=%r weighted=%r" % (counts, weighted))
    mask = counts == 0
    average = np.zeros(counts.shape, dtype=np.float64)
    average[~mask] = weighted[~mask] / counts[~mask]
    average[mask] = np.nan
    return average


class GridScope(object):
    def __init__(self, locals=None, globals=None):
        # self.locals = locals or {}
        # if locals:
        # self.__dict__.update(locals)
        self.globals = globals or {}
        self.lazy = {}
        self.lazy["average"] = grid_average
        self.globals["cumulative"] = self.cumulative
        self.globals["normalize"] = self.normalize
        self.globals.update(functions)
        self.user_added = set()

    def cumulative(self, array, normalize=True):
        mask = (np.isnan(array) | np.isinf(array))
        values = array * 1
        values[mask] = 0
        c = np.cumsum(values)
        if normalize:
            return c / c[-1]
        else:
            return c

    def normalize(self, array):
        mask = (np.isnan(array) | np.isinf(array))
        values = array * 1
        total = np.sum(values[~mask])
        return values / total

    def setter(self, key):
        def apply(value, key=key):
            self[key] = value
            return value
        return apply

    def add_lazy(self, key, f):
        self.lazy[key] = f

    def __contains__(self, item):
        return (item in self.__dict__) or (item in self.lazy)

    def __setitem__(self, key, value):
        # logger.debug("%r.__setitem__(%r, %r)" % (self, key, value))
        self.__dict__[key] = value
        self.user_added.add(key)

    def __getitem__(self, key):
        logger.debug("%r.__getitem__(%r)" % (self, key))
        if key in self.lazy and key not in self.__dict__:
            logger.debug("%r lazy item %r" % (self, key))
            f = self.lazy[key]
            self.__dict__[key] = f(self)

        return self.__dict__[key]

    def evaluate(self, expression):
        if 0:
            locals = dict(self.__dict__)
            del locals["globals"]
            logger.debug("evaluating: %r locals=%r", expression, locals)
        return eval(expression, self.globals, self)

    def slice(self, slice):
        gridscope = GridScope(globals=self.globals)
        for key in self.user_added:
            value = self[key]
            if isinstance(value, np.ndarray):
                grid = value
                sliced = np.sum(grid[slice, ...], axis=0)
                logger.debug("sliced %s from %r to %r", key, grid.shape, sliced.shape)
                gridscope[key] = sliced
            else:
                gridscope[key] = value
        return gridscope

    def disjoined(self):
        gridscope = GridScope(globals=self.globals)
        for key in self.user_added:
            value = self[key]
            if isinstance(value, np.ndarray):
                grid = vaex.utils.disjoined(value)
                gridscope[key] = grid
            else:
                gridscope[key] = value
        return gridscope

    def marginal2d(self, i, j):
        gridscope = GridScope(globals=self.globals)
        for key in self.user_added:
            value = self[key]
            if isinstance(value, np.ndarray):
                dimension = len(value.shape)
                axes = list(range(dimension))
                axes.remove(i)
                axes.remove(j)
                grid = vaex.utils.multisum(value, axes)
                gridscope[key] = grid
            else:
                gridscope[key] = value
        return gridscope


def add_mem(bytes, *info):
    global total_bytes
    total_bytes += bytes
    added = filesize_format(bytes)
    total = filesize_format(total_bytes)
    print(("MEMORY USAGE: added %s, total %s (%r)" % (added, total, info)))



def buffer(i, n=1):
    values = []
    try:
        for _ in range(n-1):
            values.append(next(i))
        while True:
            values.append(next(i))
            yield values.pop(0)
    except StopIteration:
        pass
    yield from values


def pmap(f, i, pool):
    for item in i:
        def call(args):
            return f(*args)
        yield pool.submit(call, item)


def pwait(i):
    for item in i:
        yield item.result()


def consume(i):
    for item in i:
        pass


def filter_none(i):
    for item in i:
        if item is not None:
            yield item


def chunked(i, count):
    '''Yield list 'subslices' of iterator i with max length count.

    >>> list(chunked(range(10), 2))
    [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
    >>> list(chunked(range(5), 2))
    [[0, 1], [2, 3], [4]]
    >>> list(chunked(range(3), 4))
    [[0, 1, 2]]
    >>> list(chunked(range(4), 4))
    [[0, 1, 2, 3]]
    '''
    values = []
    for el in i:
        values.append(el)
        if len(values) == count:
            yield values
            values = []
    if values:
        yield values



"""Conversion between different types of arrays"""
import numpy as np
import pyarrow as pa
import vaex.utils

supported_arrow_array_types = (pa.Array, pa.ChunkedArray)
supported_array_types = (np.ndarray, ) + supported_arrow_array_types
string_types = [pa.string(), pa.large_string()]
_type_names_int = ["int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64"]
_type_names = ["float64", "float32"] + _type_names_int
map_arrow_to_numpy = {getattr(pa, name)(): np.dtype(name) for name in _type_names}
map_arrow_to_numpy[pa.bool_()] = np.dtype("?")
for unit in 's ms us ns'.split():
    map_arrow_to_numpy[pa.timestamp(unit)] = np.dtype(f"datetime64[{unit}]")

for unit in 's ms us ns'.split():
    map_arrow_to_numpy[pa.duration(unit)] = np.dtype(f"timedelta64[{unit}]")


def full(n, value, dtype):
    from .datatype import DataType
    dtype = DataType(dtype)
    values = np.full(n, value, dtype=dtype.numpy)
    if dtype.is_arrow:
        return pa.array(values)
    else:
        return values

def is_arrow_array(ar):
    return isinstance(ar, supported_arrow_array_types)


def is_numpy_array(ar):
    return isinstance(ar, np.ndarray)


def is_array(ar):
    return is_arrow_array(ar) or is_numpy_array(ar)


def is_scalar(x):
    return not is_array(x) or (is_numpy_array(x) and x.ndim == 0)

def filter(ar, boolean_mask):
    if isinstance(ar, supported_arrow_array_types):
        return ar.filter(pa.array(boolean_mask))
    else:
        return ar[boolean_mask]


def take(ar, indices):
    return ar.take(indices)


def slice(ar, offset, length=None):
    if offset == 0 and len(ar) == length:
        return ar
    if isinstance(ar, supported_arrow_array_types):
        return ar.slice(offset, length)
    else:
        if length is not None:
            return ar[offset:offset + length]
        else:
            return ar[offset:]


def shape(ar):
    if is_arrow_array(ar):
        return (len(ar),)
    else:
        return ar.shape

def ndim(ar):
    if is_arrow_array(ar):
        return 1
    else:
        return ar.ndim


def getitem(ar, item):
    if is_arrow_array(ar):
        assert len(item) == 1, "For arrow we only support 1 d items"
        assert item[0].step is None, "Step not supported for arrow"
        start, end = item[0].start, item[0].stop
        return ar[start:end]
    else:
        return ar[item]

def concat(arrays):
    if len(arrays) == 1:
        return arrays[0]
    if any([isinstance(k, vaex.array_types.supported_arrow_array_types) for k in arrays]):
        arrays = [to_arrow(k) for k in arrays]
        flat_chunks = []
        type = arrays[0].type
        for chunk in arrays:
            if len(chunk) == 0:
                continue
            if isinstance(chunk, pa.ChunkedArray):
                flat_chunks.extend(chunk.chunks)
            else:
                flat_chunks.append(chunk)
        return pa.chunked_array(flat_chunks, type=type)

        return pa.chunked_array(arrays)
    else:
        ar = np.ma.concatenate(arrays)
        # avoid useless masks
        if ar.mask is False:
            ar = ar.data
        if ar.mask is np.False_:
            ar = ar.data
        return ar


def is_string_type(data_type):
    return not isinstance(data_type, np.dtype) and data_type in string_types


def is_string(ar):
    return isinstance(ar, supported_arrow_array_types) and is_string_type(ar.type)


def filter(ar, boolean_mask):
    if isinstance(ar, supported_arrow_array_types):
        return ar.filter(to_arrow(boolean_mask))
    else:
        return ar[boolean_mask]


def same_type(type1, type2):
    try:
        return type1 == type2
    except TypeError:
        # numpy dtypes don't like to be compared
        return False


def tolist(ar):
    if isinstance(ar, list):
        return ar
    if isinstance(ar, supported_arrow_array_types):
        return ar.to_pylist()
    else:
        return ar.tolist()


def data_type(ar):
    if isinstance(ar, supported_arrow_array_types):
        return ar.type
    else:
        return ar.dtype


def to_numpy(x, strict=False):
    import vaex.arrow.convert
    import vaex.column
    if isinstance(x, vaex.column.ColumnStringArrow):
        if strict:
            return x.to_numpy()
        else:
            return x.to_arrow()
    elif isinstance(x, np.ndarray):
        return x
    elif isinstance(x, supported_arrow_array_types):
        dtype = vaex.dtype_of(x)
        if not strict and not (dtype.is_primitive or dtype.is_temporal):
            return x
        x = vaex.arrow.convert.column_from_arrow_array(x)
        return to_numpy(x, strict=strict)
    elif hasattr(x, "to_numpy"):
        x = x.to_numpy()
    return np.asanyarray(x)


def to_arrow(x, convert_to_native=True):
    if isinstance(x, (vaex.strings.StringList32, vaex.strings.StringList64)):
        col = vaex.column.ColumnStringArrow.from_string_sequence(x)
        return pa.array(col)
    if isinstance(x, supported_arrow_array_types):
        return x
    if convert_to_native and isinstance(x, np.ndarray):
        x = vaex.utils.to_native_array(x)
    return pa.array(x)


def to_xarray(x):
    import xarray
    return xarray.DataArray(to_numpy(x))


def convert(x, type, default_type="numpy"):
    import vaex.column
    if type == "numpy":
        if isinstance(x, (list, tuple)) and len(x) > 0 and is_array(x[0]):
            return concat([convert(k, type) for k in x])
        else:
            return to_numpy(x, strict=True)
    if type == "numpy-arrow":  # used internally, numpy if possible, otherwise arrow
        if isinstance(x, (list, tuple)) and len(x) > 0 and is_array(x[0]):
            return concat([convert(k, type) for k in x])
        else:
            return to_numpy(x, strict=False)
    elif type == "arrow":
        if isinstance(x, (list, tuple)) and len(x) > 0 and is_array(x[0]):
            chunks = [convert(k, type) for k in x]
            return concat(chunks)
        else:
            return to_arrow(x)
    elif type == "xarray":
        return to_xarray(x)
    elif type in ['list', 'python']:
        if isinstance(x, (list, tuple)):
            result = []
            for chunk in x:
                result += convert(chunk, type, default_type=default_type)
            return result
        else:
            try:
                return pa.array(x).tolist()
            except:
                return np.array(x).tolist()
    elif type is None:
        if isinstance(x, (list, tuple)):
            chunks = [convert(k, type) for k in x]
            if isinstance(chunks[0], (pa.Array, pa.ChunkedArray, vaex.column.ColumnStringArrow)):
                return convert(chunks, "arrow")
            elif isinstance(chunks[0], np.ndarray):
                return convert(chunks, "numpy")
            else:
                raise ValueError("Unknown type: %r" % chunks[0])
        else:
            # return convert(x, Nonedefault_type)
            return x
    else:
        raise ValueError("Unknown type: %r" % type)


def numpy_dtype(x, strict=True):
    from . import column
    if isinstance(x, column.ColumnString):
        return x.dtype
    elif isinstance(x, np.ndarray):
        return x.dtype
    elif isinstance(x, supported_arrow_array_types):
        arrow_type = x.type
        from .datatype import DataType
        # dtype = DataType(arrow_type)
        if pa.types.is_timestamp(arrow_type):
            # https://arrow.apache.org/docs/python/pandas.html#type-differences says:
            #  'Also datetime64 is currently fixed to nanosecond resolution.'
            # so we need to do this ourselves
            unit = arrow_type.unit
            dtype = np.dtype(f'datetime64[{unit}]')
        else:
            try:
                dtype = arrow_type.to_pandas_dtype()
            except NotImplementedError:
                # assume dtype object as fallback in case arrow has no pandas dtype equivalence
                dtype = 'O'
            dtype = np.dtype(dtype)  # turn into instance
        if strict:
            return dtype
        else:
            if dtype.kind in 'iufbMm':
                return dtype
            else:
                return arrow_type

        # I don't there is a reason anymore to return this type, the to_pandas_dtype should
        # handle that
        # if isinstance(arrow_type, pa.DictionaryType):
        #     # we're interested in the type of the dictionary or the indices?
        #     if isinstance(x, pa.ChunkedArray):
        #         # take the first dictionary
        #         x = x.chunks[0]
        #     return numpy_dtype(x.dictionary)
        # if arrow_type in string_types:
        #     return arrow_type
    else:
        raise TypeError("Cannot determine numpy dtype from: %r" % x)


def arrow_type(x):
    if isinstance(x, supported_arrow_array_types):
        return x.type
    else:
        return to_arrow(x[0:1]).type


def to_arrow_type(data_type):
    data_type = vaex.dtype(data_type).internal
    if isinstance(data_type, np.dtype):
        return arrow_type_from_numpy_dtype(data_type)
    else:
        return data_type


def to_numpy_type(data_type, strict=True):
    """

    Examples:
    >>> to_numpy_type(np.dtype('f8'))
    dtype('float64')
    >>> to_numpy_type(pa.float64())
    dtype('float64')
    >>> to_numpy_type(pa.string())
    dtype('O')
    >>> to_numpy_type(pa.string(), strict=False)
    DataType(string)
    """
    if isinstance(data_type, np.dtype):
        return data_type
    else:
        return numpy_dtype_from_arrow_type(data_type, strict=strict)


def arrow_type_from_numpy_dtype(dtype):
    data = np.empty(1, dtype=dtype)
    return arrow_type(data)


def numpy_dtype_from_arrow_type(arrow_type, strict=True):
    if is_string_type(arrow_type):
        if strict:
            return np.dtype('object')
        else:
            return arrow_type
    try:
        return map_arrow_to_numpy[arrow_type]
    except KeyError:
        raise NotImplementedError(f'Cannot convert {arrow_type}')



def type_promote(t1, t2):
    # when two ndarrays, we keep it like it
    if isinstance(t1, np.dtype) and isinstance(t2, np.dtype):
        return np.promote_types(t1, t2)
    # otherwise we go to arrow
    t1 = to_arrow_type(t1)
    t2 = to_arrow_type(t2)

    if pa.types.is_null(t1):
        return t2
    if pa.types.is_null(t2):
        return t1

    if t1 == t2:
        return t1


    # TODO: so far we only use this in in code that converts to arrow
    # if we want to support numpy, we have to check it types were numpy types
    is_numerics = [pa.types.is_floating, pa.types.is_integer]
    if (any(test(t1) for test in is_numerics) and any(test(t2) for test in is_numerics)) \
       or (pa.types.is_timestamp(t1) and pa.types.is_timestamp(t2)):
        # leverage numpy for type promotion
        dtype1 = numpy_dtype_from_arrow_type(t1)
        dtype2 = numpy_dtype_from_arrow_type(t2)
        dtype = np.promote_types(dtype1, dtype2)
        return arrow_type_from_numpy_dtype(dtype)
    elif is_string_type(t1):
        return t1
    elif is_string_type(t2):
        return t2
    else:
        raise TypeError(f'Cannot promote {t1} and {t2} to a common type')


def upcast(type):
    if isinstance(type, np.dtype):
        if type.kind == "b":
            return np.dtype('int64')
        if type.kind == "i":
            return np.dtype('int64')
        if type.kind == "u":
            return np.dtype('uint64')
        if type.kind == "f":
            return np.dtype('float64')
    else:
        dtype = numpy_dtype_from_arrow_type(type)
        dtype = upcast(dtype)
        type = arrow_type_from_numpy_dtype(dtype)

    return type


def arrow_reduce_large(arrow_array):
    if arrow_array.type == pa.large_string():
        import vaex.arrow.convert
        return vaex.arrow.convert.large_string_to_string(arrow_array)
    return arrow_array



import numpy as np
import threading
import logging
import warnings

import vaex.serialize
from .expression import FunctionSerializableJit
from . import expresso



logger = logging.getLogger("vaex.webserver.tornado")


class ExpressionStringMetal(expresso.ExpressionString):
    def pow(self, left, right):
        return "pow({left}, {right})".format(left=left, right=right)


def node_to_cpp(node, pretty=False):
    return ExpressionStringMetal(pretty=pretty).visit(node)


import faulthandler
faulthandler.enable()

@vaex.serialize.register
class FunctionSerializableMetal(FunctionSerializableJit):
    device = None
    def compile(self):
        try:
            import Metal
        except ImportError:
            logging.error("Failure to import Metal, please install pyobjc-framework-Metal")
            raise
        import objc
        dtype_out = vaex.dtype(self.return_dtype).numpy
        if dtype_out.name == "float64":
            dtype_out = np.dtype("float32")
            warnings.warn("Casting output from float64 to float32 since Metal does not support float64")
        ast_node = expresso.parse_expression(self.expression)
        cppcode = node_to_cpp(ast_node)
        typemap = {'float32': 'float',
                   'float64': 'float'}  # we downcast!
        for name in vaex.array_types._type_names_int:
            typemap[name] = f'{name}_t'
        typenames = [typemap[dtype.name] for dtype in self.argument_dtypes]
        metal_args = [f'const device {typename} *{name}_array [[buffer({i})]]' for i, (typename, name) in
                      enumerate(zip(typenames, self.arguments))]
        code_get_scalar = [f'    {typename} {name} = {name}_array[id];\n' for typename, name, in zip(typenames, self.arguments)]
        sourcecode = '''
#include <metal_stdlib>
using namespace metal;

float arctan2(float y, float x) {
    return atan2(y, x);
}

template<typename T>
T where(bool condition, T y, T x) {
    return condition ? x : y;
}
kernel void vaex_kernel(%s,
                        device %s *vaex_output [[buffer(%i)]],
                        uint id [[thread_position_in_grid]]) {
%s
    vaex_output[id] = %s;
}
''' % (', '.join(metal_args), typemap[dtype_out.name], len(metal_args), ''.join(code_get_scalar), cppcode) # typemap[self.dtype_out],
        if self.verbose:
            print('Generated code:\n' + sourcecode)
        with open('test.metal', 'w') as f:
            print(f'Write to {f.name}')
            f.write(sourcecode)


        storage = threading.local()
        lock = threading.Lock()

        # following https://developer.apple.com/documentation/metal/basic_tasks_and_concepts/performing_calculations_on_a_gpu?language=objc
        self.device = Metal.MTLCreateSystemDefaultDevice()
        opts = Metal.MTLCompileOptions.new()
        self.library = self.device.newLibraryWithSource_options_error_(sourcecode, opts, objc.NULL)
        if self.library[0] is None:
            msg = f"Error compiling: {sourcecode}, sourcecode"
            logger.error(msg)
            raise RuntimeError(msg)
        kernel_name = "vaex_kernel"
        self.vaex_kernel = self.library[0].newFunctionWithName_(kernel_name)
        desc = Metal.MTLComputePipelineDescriptor.new()
        desc.setComputeFunction_(self.vaex_kernel)
        state = self.device.newComputePipelineStateWithDescriptor_error_(desc, objc.NULL)
        command_queue = self.device.newCommandQueue()

        def wrapper(*args):
            args = [vaex.array_types.to_numpy(ar) for ar in args]
            def getbuf(name, value=None, dtype=np.dtype("float32"), N=None):
                buf = getattr(storage, name, None)
                if value is not None:
                    N = len(value)
                    dtype = value.dtype
                if dtype.name == "float64":
                    warnings.warn("Casting input argument from float64 to float32 since Metal does not support float64")
                    dtype = np.dtype("float32")
                nbytes = N * dtype.itemsize
                if buf is not None and buf.length() != nbytes:
                    # doesn't match size, create a new one
                    buf = None
                # create a buffer
                if buf is None:
                    buf = self.device.newBufferWithLength_options_(nbytes, 0)
                    setattr(storage, name, buf)
                # copy data to buffer
                if value is not None:
                    mv = buf.contents().as_buffer(buf.length())
                    buf_as_numpy = np.frombuffer(mv, dtype=dtype)
                    buf_as_numpy[:] = value.astype(dtype, copy=False)
                return buf
            input_buffers = [getbuf(name, chunk) for name, chunk in zip(self.arguments, args)]
            output_buffer = getbuf('vaex_output', N=len(args[0]), dtype=dtype_out)
            buffers = input_buffers + [output_buffer]
            command_buffer = command_queue.commandBuffer()
            encoder = command_buffer.computeCommandEncoder()
            encoder.setComputePipelineState_(state)
            for i, buf in enumerate(buffers):
                encoder.setBuffer_offset_atIndex_(buf, 0, i)
            nitems = len(args[0])
            tpgrid = Metal.MTLSize(width=nitems, height=1, depth=1)
            # state.threadExecutionWidth() == 32 on M1 max
            # state.maxTotalThreadsPerThreadgroup() == 1024 on M1 max
            tptgroup = Metal.MTLSize(width=state.threadExecutionWidth(), height=state.maxTotalThreadsPerThreadgroup()//state.threadExecutionWidth(), depth=1)
            # this is simpler, and gives the same performance
            # tptgroup = Metal.MTLSize(width=1, height=1, depth=1)
            encoder.dispatchThreads_threadsPerThreadgroup_(tpgrid, tptgroup)
            encoder.endEncoding()
            command_buffer.commit()
            command_buffer.waitUntilCompleted()

            output_buffer_py = output_buffer.contents().as_buffer(output_buffer.length())
             # do we needs .copy() ?
            result = np.frombuffer(output_buffer_py, dtype=dtype_out)
            return result
        return wrapper



import os
import math
import threading

import pkg_resources
import vaex.settings

# thread local variable, where 'global' tracker go
local = threading.local()
_memory_tracker_types = {}
lock = threading.Lock()



class MemoryTracker:
    track_live = False
    def __init__(self) -> None:
        self.used = 0

    def pre_alloc(self, bytes, reason):
        self.used += bytes

    def using(self, bytes):
        self.used += bytes


def create_tracker():
    memory_tracker_type = vaex.settings.main.memory_tracker.type
    if not _memory_tracker_types:
        with lock:
            if not _memory_tracker_types:
                for entry in pkg_resources.iter_entry_points(group="vaex.memory.tracker"):
                    _memory_tracker_types[entry.name] = entry.load()
    cls = _memory_tracker_types.get(memory_tracker_type)
    if cls is not None:
        return cls()
    raise ValueError(f"No memory tracker found with name {memory_tracker_type}")
